@@ -27,6 +27,7 @@ const MESSAGE: mio::Token = mio::Token(1);
 // -  messages telling it what to do
 //
 // -  events telling it that something has happened on one of the watched files.
+
 struct EventLoop {
     running: bool,
     poll: mio::Poll,
@@ -35,7 +36,8 @@ struct EventLoop {
     event_loop_rx: Receiver<EventLoopMsg>,
     inotify: Option<Inotify>,
     event_handler: Box<dyn EventHandler>,
-    watches: HashMap<PathBuf, (WatchDescriptor, WatchMask, bool)>,
+    /// PathBuf -> (WatchDescriptor, WatchMask, is_recursive, is_dir)
+    watches: HashMap<PathBuf, (WatchDescriptor, WatchMask, bool, bool)>,
     paths: HashMap<WatchDescriptor, PathBuf>,
     rename_event: Option<Event>,
 }
@@ -58,13 +60,13 @@ enum EventLoopMsg {
 fn add_watch_by_event(
     path: &Option<PathBuf>,
     event: &inotify_sys::Event<&OsStr>,
-    watches: &HashMap<PathBuf, (WatchDescriptor, WatchMask, bool)>,
+    watches: &HashMap<PathBuf, (WatchDescriptor, WatchMask, bool, bool)>,
     add_watches: &mut Vec<PathBuf>,
 ) {
     if let Some(ref path) = *path {
         if event.mask.contains(EventMask::ISDIR) {
             if let Some(parent_path) = path.parent() {
-                if let Some(&(_, _, is_recursive)) = watches.get(parent_path) {
+                if let Some(&(_, _, is_recursive, _)) = watches.get(parent_path) {
                     if is_recursive {
                         add_watches.push(path.to_owned());
                     }
@@ -77,7 +79,7 @@ fn add_watch_by_event(
 #[inline]
 fn remove_watch_by_event(
     path: &Option<PathBuf>,
-    watches: &HashMap<PathBuf, (WatchDescriptor, WatchMask, bool)>,
+    watches: &HashMap<PathBuf, (WatchDescriptor, WatchMask, bool, bool)>,
     remove_watches: &mut Vec<PathBuf>,
 ) {
     if let Some(ref path) = *path {
@@ -202,6 +204,8 @@ impl EventLoop {
                     Ok(events) => {
                         let mut num_events = 0;
                         for event in events {
+                            log::trace!("inotify event: {event:?}");
+
                             num_events += 1;
                             if event.mask.contains(EventMask::Q_OVERFLOW) {
                                 let ev = Ok(Event::new(EventKind::Other).set_flag(Flag::Rescan));
@@ -209,9 +213,7 @@ impl EventLoop {
                             }
 
                             let path = match event.name {
-                                Some(name) => {
-                                    self.paths.get(&event.wd).map(|root| root.join(&name))
-                                }
+                                Some(name) => self.paths.get(&event.wd).map(|root| root.join(name)),
                                 None => self.paths.get(&event.wd).cloned(),
                             };
 
@@ -254,14 +256,8 @@ impl EventLoop {
                                         .add_some_path(rename_event.paths.first().cloned())
                                         .add_some_path(path.clone()),
                                     );
-                                } else {
-                                    add_watch_by_event(
-                                        &path,
-                                        &event,
-                                        &self.watches,
-                                        &mut add_watches,
-                                    );
                                 }
+                                add_watch_by_event(&path, &event, &self.watches, &mut add_watches);
                             }
                             if event.mask.contains(EventMask::MOVE_SELF) {
                                 evs.push(
@@ -287,9 +283,7 @@ impl EventLoop {
                                 );
                                 add_watch_by_event(&path, &event, &self.watches, &mut add_watches);
                             }
-                            if event.mask.contains(EventMask::DELETE_SELF)
-                                || event.mask.contains(EventMask::DELETE)
-                            {
+                            if event.mask.contains(EventMask::DELETE) {
                                 evs.push(
                                     Event::new(EventKind::Remove(
                                         if event.mask.contains(EventMask::ISDIR) {
@@ -302,6 +296,29 @@ impl EventLoop {
                                 );
                                 remove_watch_by_event(&path, &self.watches, &mut remove_watches);
                             }
+                            if event.mask.contains(EventMask::DELETE_SELF) {
+                                let remove_kind = match &path {
+                                    Some(watched_path) => {
+                                        let current_watch = self.watches.get(watched_path);
+                                        match current_watch {
+                                            Some(&(_, _, _, true)) => RemoveKind::Folder,
+                                            Some(&(_, _, _, false)) => RemoveKind::File,
+                                            None => RemoveKind::Other,
+                                        }
+                                    }
+                                    None => {
+                                        log::trace!(
+                                            "No patch for DELETE_SELF event, may be a bug?"
+                                        );
+                                        RemoveKind::Other
+                                    }
+                                };
+                                evs.push(
+                                    Event::new(EventKind::Remove(remove_kind))
+                                        .add_some_path(path.clone()),
+                                );
+                                remove_watch_by_event(&path, &self.watches, &mut remove_watches);
+                            }
                             if event.mask.contains(EventMask::MODIFY) {
                                 evs.push(
                                     Event::new(EventKind::Modify(ModifyKind::Data(
@@ -353,6 +370,10 @@ impl EventLoop {
                             break;
                         }
                     }
+                    Err(e) if e.kind() == std::io::ErrorKind::WouldBlock => {
+                        // No events read. Break out.
+                        break;
+                    }
                     Err(e) => {
                         self.event_handler.handle_event(Err(Error::io(e)));
                     }
@@ -372,7 +393,7 @@ impl EventLoop {
     fn add_watch(&mut self, path: PathBuf, is_recursive: bool, mut watch_self: bool) -> Result<()> {
         // If the watch is not recursive, or if we determine (by stat'ing the path to get its
         // metadata) that the watched path is not a directory, add a single path watch.
-        if !is_recursive || !metadata(&path).map_err(Error::io)?.is_dir() {
+        if !is_recursive || !metadata(&path).map_err(Error::io_watch)?.is_dir() {
             return self.add_single_watch(path, false, true);
         }
 
@@ -396,6 +417,7 @@ impl EventLoop {
     ) -> Result<()> {
         let mut watchmask = WatchMask::ATTRIB
             | WatchMask::CREATE
+            | WatchMask::OPEN
             | WatchMask::DELETE
             | WatchMask::CLOSE_WRITE
             | WatchMask::MODIFY
@@ -407,13 +429,15 @@ impl EventLoop {
             watchmask.insert(WatchMask::MOVE_SELF);
         }
 
-        if let Some(&(_, old_watchmask, _)) = self.watches.get(&path) {
+        if let Some(&(_, old_watchmask, _, _)) = self.watches.get(&path) {
             watchmask.insert(old_watchmask);
             watchmask.insert(WatchMask::MASK_ADD);
         }
 
         if let Some(ref mut inotify) = self.inotify {
-            match inotify.add_watch(&path, watchmask) {
+            log::trace!("adding inotify watch: {}", path.display());
+
+            match inotify.watches().add(&path, watchmask) {
                 Err(e) => {
                     Err(if e.raw_os_error() == Some(libc::ENOSPC) {
                         // do not report inotify limits as "no more space" on linux #266
@@ -425,8 +449,9 @@ impl EventLoop {
                 }
                 Ok(w) => {
                     watchmask.remove(WatchMask::MASK_ADD);
+                    let is_dir = metadata(&path).map_err(Error::io)?.is_dir();
                     self.watches
-                        .insert(path.clone(), (w.clone(), watchmask, is_recursive));
+                        .insert(path.clone(), (w.clone(), watchmask, is_recursive, is_dir));
                     self.paths.insert(w, path);
                     Ok(())
                 }
@@ -439,10 +464,13 @@ impl EventLoop {
     fn remove_watch(&mut self, path: PathBuf, remove_recursive: bool) -> Result<()> {
         match self.watches.remove(&path) {
             None => return Err(Error::watch_not_found().add_path(path)),
-            Some((w, _, is_recursive)) => {
+            Some((w, _, is_recursive, _)) => {
                 if let Some(ref mut inotify) = self.inotify {
-                    inotify
-                        .rm_watch(w.clone())
+                    let mut inotify_watches = inotify.watches();
+                    log::trace!("removing inotify watch: {}", path.display());
+
+                    inotify_watches
+                        .remove(w.clone())
                         .map_err(|e| Error::io(e).add_path(path.clone()))?;
                     self.paths.remove(&w);
 
@@ -450,8 +478,8 @@ impl EventLoop {
                         let mut remove_list = Vec::new();
                         for (w, p) in &self.paths {
                             if p.starts_with(&path) {
-                                inotify
-                                    .rm_watch(w.clone())
+                                inotify_watches
+                                    .remove(w.clone())
                                     .map_err(|e| Error::io(e).add_path(p.into()))?;
                                 self.watches.remove(p);
                                 remove_list.push(w.clone());
@@ -469,9 +497,10 @@ impl EventLoop {
 
     fn remove_all_watches(&mut self) -> Result<()> {
         if let Some(ref mut inotify) = self.inotify {
+            let mut inotify_watches = inotify.watches();
             for (w, p) in &self.paths {
-                inotify
-                    .rm_watch(w.clone())
+                inotify_watches
+                    .remove(w.clone())
                     .map_err(|e| Error::io(e).add_path(p.into()))?;
             }
             self.watches.clear();
